from .file_box import FileBox

from .type import (
    FileBoxType,
    FileBoxOptionsQrCode,
    FileBoxOptionsBase64,
    FileBoxOptionsBuffer,
    FileBoxOptionsStream,
    FileBoxOptionsBase,
    FileBoxOptionsFile,
    FileBoxOptionsUrl
)

__all__ = [
    'FileBox',
    'FileBoxOptionsUrl',
    'FileBoxOptionsFile',
    'FileBoxOptionsBase',
    'FileBoxOptionsStream',
    'FileBoxOptionsBuffer',
    'FileBoxOptionsBase64',
    'FileBoxOptionsQrCode',
    'FileBoxType'
]



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from dataclasses import Field, dataclass, fields, is_dataclass, MISSING
import importlib
import os

from typing import Any, Dict, List, Optional, Tuple

from wechaty_puppet.logger import get_logger

logger = get_logger('WechatyPuppet.Schema.Base')


def _get_verison_wechaty_related_package(module_name: str) -> Optional[str]:
    """
    get the version of wechaty related package
    """
    if not module_name.endswith('.version'):
        module_name += '.version'
    try:
        module = importlib.import_module(module_name)
        version = None
        if hasattr(module, 'VERSION'):
            version = getattr(module, 'VERSION')
        return version
    except ImportError:
        return None


def _guess_token_type() -> str:
    token = os.environ.get('WECHATY_PUPPET_SERVICE_TOKEN', None) or \
        os.environ.get('TOKEN', None) or \
        os.environ.get('token', None) or None
    unknown_token = '<unkonwn> token'
    if not token:
        return unknown_token
    
    if token.startswith('puppet_padlocal_'):
        return 'padlocal'
    if token.startswith('puppet_wxwork'):
        return 'wxwork'
    return unknown_token


@dataclass
class BaseDataClass:
    """Abstract Event Payload which handle initialization"""

    def __init__(self, **kwargs) -> None:

        all_fields = fields(self)
        invalid_fields = self.check_fields(kwargs)
        if invalid_fields:
            BaseDataClass._print_warning_about_dataclass_fields(self, kwargs)

            for invalid_field in invalid_fields:
                kwargs.pop(invalid_field)

        for field in all_fields:
            value = kwargs.get(field.name, None)
            if value is None:
                value = field.default

            if field.default_factory is not MISSING:
                value = field.default_factory()

            if value is MISSING:
                raise ValueError(
                    f'the field<{field.name}> is required in dataclass<{type(self)}>'
                )

            setattr(self, field.name, value)

    @classmethod
    def check_fields(cls, kwargs: Dict[str, Any]) -> List[str]:
        """check if the input data is valid to initialize the dataclass

        Args:
            cls (subclass of EventPayloadBase): the type of the dataclass
            kwargs (Dict[str, Any]): the input data

        Returns:
            List[str]: the invalid fields
        """
        if not is_dataclass(cls):
            raise TypeError(f'{cls} is not a subclass of EventPayloadBase')

        dataclass_fields: List[Field] = [field for field in fields(cls)]
        dataclass_field_names: set = {
            field.name for field in dataclass_fields}
        return list(set(kwargs.keys()) - dataclass_field_names)

    @staticmethod
    def _print_warning_about_dataclass_fields(obj, kwargs: Dict[str, Any]):
        """print warning info and notice that the developer
        can post issue to help python-wechaty to improve the quality of code

        Args:
            obj (_type_): the instance of the dataclass
            kwargs (Dict[str, Any]): the arguments of __init__ function 
        """
        # 1. check if the obj is a dataclass
        if not is_dataclass(obj):
            return

        # 2. get invalid fields
        invalid_fields = BaseDataClass.check_fields(kwargs)
        if not invalid_fields:
            return

        # 3. get the version of token/wechaty/wechaty-puppet/wechaty-puppet-service packages
        package_versions = {
            "wechaty": "0.0.0",
            "wechaty_puppet": "0.0.0",
            "wechaty_puppet_service": "0.0.0",
        }
        for package_module, default_version in package_versions.items():
            package_version = _get_verison_wechaty_related_package(
                package_module)
            package_versions[package_module] = package_version or default_version

        # 5. guess the token type
        token_type = _guess_token_type()

        # 4. print warning info
        logger.warning(
            f'invalid fields: <{invalid_fields}> to initialize type<{type(obj)}> '
            'which are should not be here, we suggest that you should post an issue to python-wechaty to fix this issue \n'
            'you should give these infos: \n'
            f'* the version of wechaty: \t\t\t{package_versions["wechaty"]}\n'
            f'* the version of wechaty_puppet: \t\t{package_versions["wechaty_puppet"]}\n'
            f'* the version of wechaty_puppet_service: \t{package_versions["wechaty_puppet_service"]}\n'
            f'* the token type: {token_type} \n'
        )



"""
file-box utils functions
"""
import mimetypes
import os
from typing import Tuple
from urllib.parse import urlparse
import requests
import uuid


def data_url_to_base64(data_url: str) -> str:
    """
    transfer the DataUrl to base64 format

    data_url:  dataURL: `data:image/png;base64,${base64Text}`,

    return :
    """
    if ',' not in data_url:
        raise ValueError('DataUrl value is not valid, value: %s', data_url)

    return data_url.split(',')[-1]


def _extract_content_type(content_type: str) -> str:
    """extract content type"""
    if ';' in content_type:
        content_type = content_type.split(';')[0]

    if '/' in content_type:
        content_type = content_type.split('/')[-1]
    return content_type


def extract_file_name_from_url(url: str) -> Tuple[str, str]:
    """
    """
    # if we can guess the type from url
    res = requests.get(url)
    content_type = _extract_content_type(res.headers.get('Content-Type', ''))

    if mimetypes.guess_type(url)[0]:
        url_path = urlparse(url).path

        try:
            # refer to : https://stackoverflow.com/a/18727481/6894382
            file_name = os.path.basename(url_path)
            return file_name, content_type
        except TypeError as type_error:
            # return then random name of the file
            file_type, _ = mimetypes.guess_type(url)
            assert file_type is not None
            return '{}.{}'.format(str(uuid.uuid4()), file_type.split('/')[0]), \
                   content_type
    else:

        if res.status_code != 200:
            raise FileNotFoundError('can"t get the file from url <%s>' % url)

        if '?' in url:
            url = url[:url.index('?')]

        if url.endswith('/'):
            url = url[:-1]

        suffix = ''
        if not url.endswith(f'.{content_type}'):
            # if the url contains the file-name, so return it
            suffix = f'.{content_type}'

        file_name = f'{url[url.rindex("/")+1 :]}{suffix}'
        return file_name, content_type



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from dataclasses import dataclass
from typing import Optional

from wechaty_puppet.schemas.base import BaseDataClass


@dataclass(init=False)
class MiniProgramPayload(BaseDataClass):
    """
    mini_program payload
    """
    appid: Optional[str] = None
    description: Optional[str] = None
    pagePath: Optional[str] = None
    thumbKey: Optional[str] = None
    iconUrl: Optional[str] = None
    thumbUrl: Optional[str] = None
    shareId: Optional[str] = None
    title: Optional[str] = None
    username: Optional[str] = None



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2020-now @ Copyright Wechaty

Licensed under the Apache License, Version 2.0 (the 'License');
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an 'AS IS' BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations
from enum import Enum
from typing import Optional

from dataclasses import dataclass

from wechaty_puppet.schemas.base import BaseDataClass

from .types import MessageType


class WechatAppMessageType(Enum):
    Text = 1
    Img = 2
    Audio = 3
    Video = 4
    Url = 5
    Attach = 6
    Open = 7
    Emoji = 8
    VoiceRemind = 9
    ScanGood = 10
    Good = 13
    Emotion = 15
    CardTicket = 16
    RealtimeShareLocation = 17
    ChatHistory = 19
    MiniProgram = 33
    Transfers = 2000
    RedEnvelopes = 2001
    ReaderType = 100001


class WechatMessageType(Enum):
    Text = 1
    Image = 3
    Voice = 34
    VerifyMsg = 37
    PossibleFriendMsg = 40
    ShareCard = 42
    Video = 43
    Emoticon = 47
    Location = 48
    App = 49
    VoipMsg = 50
    StatusNotify = 51
    VoipNotify = 52
    VoipInvite = 53
    MicroVideo = 62
    # 转账
    Transfer = 2000
    # 红包
    RedEnvelope = 2001
    # 小程序
    MiniProgram = 2002
    # 群邀请
    GroupInvite = 2003
    # 文件消息
    File = 2004
    SysNotice = 9999
    Sys = 10000
    # NOTIFY 服务通知
    Recalled = 10002


@dataclass(init=False)
class MessagePayloadRoom(BaseDataClass):

    room_id: str
    from_id: Optional[str] = None

    # if to is not set, then room must be set
    to_id: Optional[str] = None


@dataclass(init=False)
class MessagePayloadTo(BaseDataClass):
    from_id: str

    # if to is not set, then room must be set
    to_id: str

    room_id: Optional[str] = None


@dataclass(init=False)
class MessageQueryFilter(BaseDataClass):
    from_id: Optional[str] = None
    id: Optional[str] = None
    room_id: Optional[str] = None
    text: Optional[str] = None
    to_id: Optional[str] = None
    type: Optional[MessageType] = None



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2018-now @copyright Wechaty

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import annotations

from enum import Enum
from dataclasses import dataclass

from typing import (
    Optional,
    Any, Dict
)

from wechaty_puppet.schemas.base import BaseDataClass


Metadata = Dict[str, Any]


class FileBoxType(Enum):
    """
    option config for file box
    """
    Unknown = 0

    # Serializable by toJSON()

    Base64 = 1
    Url = 2
    QRCode = 3

    # Not serializable by toJSON()
    # Need to convert to FileBoxType.Base64 before call toJSON()
    Buffer = 4
    File = 5
    Stream = 6


@dataclass(init=False)
class FileBoxOptionsBase(BaseDataClass):
    name: str
    type: FileBoxType


@dataclass(init=False)
class FileBoxOptionsUrl(FileBoxOptionsBase):
    """
    url file-box options
    """
    url: str = ''
    headers: Optional[dict] = None
    type: FileBoxType = FileBoxType.Url


@dataclass(init=False)
class FileBoxOptionsFile(FileBoxOptionsBase):
    """
    file file-box options
    """
    path: Optional[str] = None
    type: FileBoxType = FileBoxType.File


@dataclass(init=False)
class FileBoxOptionsBuffer(FileBoxOptionsBase):
    """
    file file-box options
    """
    # TODO -> recheck the buffer data type
    buffer: bytes = b''
    type: FileBoxType = FileBoxType.Buffer


@dataclass(init=False)
class FileBoxOptionsStream(FileBoxOptionsBase):
    """
    stream file-box options
    """
    # TODO -> recheck the stream data type
    stream: bytes = b''
    type: FileBoxType = FileBoxType.Stream


@dataclass(init=False)
class FileBoxOptionsQrCode(FileBoxOptionsBase):
    """
    qr-code file-box options
    """
    qr_code: str = ''
    type: FileBoxType = FileBoxType.QRCode


@dataclass(init=False)
class FileBoxOptionsBase64(FileBoxOptionsBase):
    """
    base64 file-box options
    """
    base64: bytes = b''
    type: FileBoxType = FileBoxType.Base64



"""
Python Wechaty - https://github.com/wechaty/python-wechaty

Authors:    Huan LI (李卓桓) <https://github.com/huan>
            Jingjing WU (吴京京) <https://github.com/wj-Mcat>

2018-now @copyright Wechaty

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
from __future__ import annotations

from enum import Enum
from typing import Optional, Union


class StateType(Enum):
    """
    state type
    """
    On = 0
    Off = 1
    Pending = 2


class StateSwitch:
    """
    puppet state switcher
    """
    def __init__(self, name: Optional[str] = None):
        self.name: Optional[str] = name
        self._state = StateType.Off

    def on(self, state: Optional[Union[bool, StateType]] = None
           ) -> Union[bool, StateType]:
        """
        Turn on the current state.
        """
        if state is None:
            if self._state == StateType.On:
                return True
            return self._state

        if isinstance(state, bool):
            if not state:
                raise ValueError(
                    'when set on state, bool value should be True')

            self._state = StateType.On
            return True
        if isinstance(state, StateType):
            if state != StateType.Pending:
                raise ValueError(
                    f'state StateType value <{state}> should be Pending')
            self._state = StateType.Pending
            return state
        raise ValueError(
            'state params type is only for [bool, StateType]')

    def off(self, state: Optional[Union[bool, StateType]] = None
            ) -> Union[bool, StateType]:
        """
        Turn Off the current state
        """
        if state is None:
            if self._state == StateType.Off:
                return True
            return self._state

        if isinstance(state, bool):
            if not state:
                raise ValueError(
                    'when set off state, bool value should be False')
            self._state = StateType.On
            return True
        if isinstance(state, StateType):
            if state != StateType.Pending:
                raise ValueError(
                    f'state StateType value <{state}> should be Pending')
            self._state = StateType.Pending
            return state
        raise ValueError(
            'state params type is only for [bool, StateType]')
